from pydantic import BaseModel, ConfigDict

from zrb.group.any_group import AnyGroup
from zrb.task.any_task import AnyTask
from zrb.util.group import get_all_subtasks


class User(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    username: str
    password: str = ""
    is_super_admin: bool = False
    is_guest: bool = False
    accessible_tasks: list[AnyTask | str] = []

    def is_password_match(self, password: str) -> bool:
        return self.password == password

    def can_access_group(self, group: AnyGroup) -> bool:
        if self.is_super_admin:
            return True
        all_tasks = get_all_subtasks(group, web_only=True)
        if any(self.can_access_task(task) for task in all_tasks):
            return True
        return False

    def can_access_task(self, task: AnyTask) -> bool:
        if self.is_super_admin:
            return True
        if task.name in self.accessible_tasks or task in self.accessible_tasks:
            return True
        return False



from abc import ABC, abstractmethod

from my_app_name.schema.permission import (
    Permission,
    PermissionCreateWithAudit,
    PermissionResponse,
    PermissionUpdateWithAudit,
)


class PermissionRepository(ABC):

    @abstractmethod
    async def get_by_id(self, id: str) -> PermissionResponse:
        """Get permission by id"""

    @abstractmethod
    async def get_by_ids(self, id_list: list[str]) -> PermissionResponse:
        """Get permissions by ids"""

    @abstractmethod
    async def get(
        self,
        page: int = 1,
        page_size: int = 10,
        filter: str | None = None,
        sort: str | None = None,
    ) -> list[Permission]:
        """Get permissions by filter and sort"""

    @abstractmethod
    async def count(self, filter: str | None = None) -> int:
        """Count permissions by filter"""

    @abstractmethod
    async def create(self, data: PermissionCreateWithAudit) -> Permission:
        """Create a new permission"""

    @abstractmethod
    async def create_bulk(
        self, data: list[PermissionCreateWithAudit]
    ) -> list[Permission]:
        """Create some permissions"""

    @abstractmethod
    async def delete(self, id: str) -> Permission:
        """Delete a permission"""

    @abstractmethod
    async def delete_bulk(self, id_list: list[str]) -> list[Permission]:
        """Delete some permissions"""

    @abstractmethod
    async def update(self, id: str, data: PermissionUpdateWithAudit) -> Permission:
        """Update a permission"""

    @abstractmethod
    async def update_bulk(
        self, id_list: list[str], data: PermissionUpdateWithAudit
    ) -> list[Permission]:
        """Update some permissions"""



"""
This module provides functions for managing interactive chat sessions with an LLM.

It handles reading user input, triggering the LLM task, and managing the
conversation flow via XCom.
"""

import asyncio

from zrb.context.any_context import AnyContext
from zrb.util.cli.style import stylize_bold_yellow, stylize_faint


async def read_user_prompt(ctx: AnyContext) -> str:
    """
    Reads user input from the CLI for an interactive chat session.

    Handles special commands like /bye, /multi, /end, and /help.
    Triggers the LLM task and waits for the result.

    Args:
        ctx: The context object for the task.

    Returns:
        The final result from the LLM session.
    """
    _show_info(ctx)
    final_result = ""
    ctx.print(stylize_faint("🧑 >> ") + f"{ctx.input.message}", plain=True)
    result = await _trigger_ask_and_wait_for_result(
        ctx,
        user_prompt=ctx.input.message,
        previous_session_name=ctx.input.previous_session,
        start_new=ctx.input.start_new,
    )
    if ctx.env.get("_ZRB_WEB_ENV", "0") != "0":
        # Don't run in web environment
        if result is not None:
            final_result = result
        return final_result
    multiline_mode = False
    user_inputs = []
    while True:
        await asyncio.sleep(0.01)
        ctx.print(stylize_faint("🧑 >> "), end="", plain=True)
        user_input = input()
        # Handle special input
        if user_input.strip().lower() in ("/bye", "/quit"):
            user_prompt = "\n".join(user_inputs)
            user_inputs = []
            result = await _trigger_ask_and_wait_for_result(ctx, user_prompt)
            if result is not None:
                final_result = result
            break
        elif user_input.strip().lower() in ("/multi"):
            multiline_mode = True
        elif user_input.strip().lower() in ("/end"):
            multiline_mode = False
            user_prompt = "\n".join(user_inputs)
            user_inputs = []
            result = await _trigger_ask_and_wait_for_result(ctx, user_prompt)
            if result is not None:
                final_result = result
        elif user_input.strip().lower() in ("/help", "/info"):
            _show_info(ctx)
            continue
        else:
            user_inputs.append(user_input)
            if multiline_mode:
                continue
            user_prompt = "\n".join(user_inputs)
            user_inputs = []
            result = await _trigger_ask_and_wait_for_result(ctx, user_prompt)
            if result is not None:
                final_result = result
    return final_result


async def _trigger_ask_and_wait_for_result(
    ctx: AnyContext,
    user_prompt: str,
    previous_session_name: str | None = None,
    start_new: bool = False,
) -> str | None:
    """
    Triggers the LLM ask task and waits for the result via XCom.

    Args:
        ctx: The context object for the task.
        user_prompt: The user's message to send to the LLM.
        previous_session_name: The name of the previous chat session (optional).
        start_new: Whether to start a new conversation (optional).

    Returns:
        The result from the LLM task, or None if the user prompt is empty.
    """
    if user_prompt.strip() == "":
        return None
    await _trigger_ask(ctx, user_prompt, previous_session_name, start_new)
    result = await _wait_ask_result(ctx)
    ctx.print(f"{stylize_faint('🤖 >>')} {result}", plain=True)
    return result


def get_llm_ask_input_mapping(callback_ctx: AnyContext):
    """
    Generates the input mapping for the LLM ask task from the callback context.

    Args:
        callback_ctx: The context object for the callback.

    Returns:
        A dictionary containing the input mapping for the LLM ask task.
    """
    data = callback_ctx.xcom.ask_trigger.pop()
    return {
        "model": callback_ctx.input.model,
        "base-url": callback_ctx.input.base_url,
        "api-key": callback_ctx.input.api_key,
        "system-prompt": callback_ctx.input.system_prompt,
        "start-new": data.get("start_new"),
        "previous-session": data.get("previous_session_name"),
        "message": data.get("message"),
    }


async def _trigger_ask(
    ctx: AnyContext,
    user_prompt: str,
    previous_session_name: str | None = None,
    start_new: bool = False,
):
    """
    Triggers the LLM ask task by pushing data to the 'ask_trigger' XCom queue.

    Args:
        ctx: The context object for the task.
        user_prompt: The user's message to send to the LLM.
        previous_session_name: The name of the previous chat session (optional).
        start_new: Whether to start a new conversation (optional).
    """
    if previous_session_name is None:
        previous_session_name = await _wait_ask_session_name(ctx)
    ctx.xcom["ask_trigger"].push(
        {
            "previous_session_name": previous_session_name,
            "start_new": start_new,
            "message": user_prompt,
        }
    )


async def _wait_ask_result(ctx: AnyContext) -> str | None:
    """
    Waits for and retrieves the LLM task result from the 'ask_result' XCom queue.

    Args:
        ctx: The context object for the task.

    Returns:
        The result string from the LLM task.
    """
    while "ask_result" not in ctx.xcom or len(ctx.xcom.ask_result) == 0:
        await asyncio.sleep(0.1)
        if "ask_error" in ctx.xcom and len(ctx.xcom.ask_error) > 0:
            ctx.xcom.ask_error.pop()
            return None
    return ctx.xcom.ask_result.pop()


async def _wait_ask_session_name(ctx: AnyContext) -> str:
    """
    Waits for and retrieves the LLM chat session name from the 'ask_session_name' XCom queue.

    Args:
        ctx: The context object for the task.

    Returns:
        The session name string.
    """
    while "ask_session_name" not in ctx.xcom or len(ctx.xcom.ask_session_name) == 0:
        await asyncio.sleep(0.1)
    return ctx.xcom.ask_session_name.pop()


def _show_info(ctx: AnyContext):
    """
    Displays the available chat session commands to the user.

    Args:
        ctx: The context object for the task.
    """
    ctx.print(
        "\n".join(
            [
                f"{stylize_bold_yellow('/bye')}   {stylize_faint('Quit from chat session')}",
                f"{stylize_bold_yellow('/multi')} {stylize_faint('Start multiline input')}",
                f"{stylize_bold_yellow('/end')}   {stylize_faint('End multiline input')}",
                f"{stylize_bold_yellow('/help')}  {stylize_faint('Show this message')}",
            ]
        ),
        plain=True,
    )



from typing import Any

from zrb.attr.type import StrDictAttr
from zrb.callback.any_callback import AnyCallback
from zrb.session.any_session import AnySession
from zrb.task.any_task import AnyTask
from zrb.util.attr import get_str_dict_attr
from zrb.util.string.conversion import to_snake_case
from zrb.xcom.xcom import Xcom


class Callback(AnyCallback):
    """
    Represents a callback that runs a task after a trigger or scheduler event.

    It handles input mapping and can publish results and session names
    back to the parent session via XCom.
    """

    def __init__(
        self,
        task: AnyTask,
        input_mapping: StrDictAttr,
        render_input_mapping: bool = True,
        result_queue: str | None = None,
        error_queue: str | None = None,
        session_name_queue: str | None = None,
    ):
        """
        Initializes a new instance of the Callback class.

        Args:
            task: The task to be executed by the callback.
            input_mapping: A dictionary or attribute mapping to prepare inputs for the task.
            render_input_mapping: Whether to render the input mapping using
                f-string like syntax.
            result_queue: The name of the XCom queue in the parent session
                to publish the task result.
            result_queue: The name of the Xcom queue in the parent session
                to publish the task error.
            session_name_queue: The name of the XCom queue in the parent
                session to publish the session name.
        """
        self._task = task
        self._input_mapping = input_mapping
        self._render_input_mapping = render_input_mapping
        self._result_queue = result_queue
        self._error_queue = error_queue
        self._session_name_queue = session_name_queue

    async def async_run(self, parent_session: AnySession, session: AnySession) -> Any:
        self._maybe_publish_session_name_to_parent_session(
            parent_session=parent_session, session=session
        )
        # prepare input
        inputs = get_str_dict_attr(
            session.shared_ctx,
            self._input_mapping,
            auto_render=self._render_input_mapping,
        )
        for name, value in inputs.items():
            session.shared_ctx.input[name] = value
            session.shared_ctx.input[to_snake_case(name)] = value
        # run task and get result
        try:
            result = await self._task.async_run(session)
            self._maybe_publish_result_to_parent_session(parent_session, result)
            return result
        except BaseException as e:
            self._maybe_publish_error_to_parent_session(parent_session, e)

    def _maybe_publish_session_name_to_parent_session(
        self, parent_session: AnySession, session: AnySession
    ):
        self._maybe_publish_to_parent_session(
            parent_session=parent_session,
            xcom_name=self._session_name_queue,
            value=session.name,
        )

    def _maybe_publish_error_to_parent_session(
        self, parent_session: AnySession, error: Any
    ):
        self._maybe_publish_to_parent_session(
            parent_session=parent_session, xcom_name=self._error_queue, value=error
        )

    def _maybe_publish_result_to_parent_session(
        self, parent_session: AnySession, result: Any
    ):
        self._maybe_publish_to_parent_session(
            parent_session=parent_session, xcom_name=self._result_queue, value=result
        )

    def _maybe_publish_to_parent_session(
        self, parent_session: AnySession, xcom_name: str | None, value: Any
    ):
        if xcom_name is None:
            return
        parent_xcom = parent_session.shared_ctx.xcom
        if xcom_name not in parent_xcom:
            parent_xcom[xcom_name] = Xcom([])
        parent_xcom[xcom_name].push(value)



from typing import Any

from zrb.builtin.group import http_group
from zrb.context.any_context import AnyContext
from zrb.input.option_input import OptionInput
from zrb.input.str_input import StrInput
from zrb.task.make_task import make_task


@make_task(
    name="http-request",
    description="🌐 Send HTTP request (Postman-like)",
    group=http_group,
    alias="request",
    input=[
        OptionInput(
            name="method",
            default="GET",
            options=["GET", "POST", "PUT", "DELETE", "PATCH"],
            prompt="HTTP method",
        ),
        StrInput(name="url", description="Target URL", prompt="Enter request URL"),
        StrInput(
            name="headers",
            description="Request headers (JSON)",
            prompt="Enter headers as JSON",
            default="{}",
        ),
        StrInput(
            name="body",
            description="Request body (JSON)",
            prompt="Enter body as JSON",
            default="{}",
        ),
        OptionInput(
            name="verify_ssl",
            default="true",
            options=["true", "false"],
            description="Verify SSL certificate",
        ),
    ],
)
def http_request(ctx: AnyContext) -> Any:
    import json

    import requests

    try:
        # Prepare headers
        headers = json.loads(ctx.input.headers)

        # Prepare body
        body = None
        if ctx.input.body != "{}":
            body = json.loads(ctx.input.body)

        # Make request
        verify = ctx.input.verify_ssl.lower() == "true"
        response = requests.request(
            method=ctx.input.method,
            url=ctx.input.url,
            headers=headers,
            json=body,
            verify=verify,
        )

        # Print request/response details
        ctx.print("🌐 Request:")
        ctx.print(f"  Method: {ctx.input.method}")
        ctx.print(f"  URL: {ctx.input.url}")
        ctx.print(f"  Headers: {headers}")
        ctx.print(f"  Body: {body}")
        ctx.print(f"  Verify SSL: {verify}")
        ctx.print("📥 Response:")
        ctx.print(f"  Status: {response.status_code}")
        ctx.print(f"  Headers: {dict(response.headers)}")
        ctx.print(f"  Body: {response.text}")

        return response
    except Exception as e:
        ctx.print_err(f"HTTP request failed: {e}")
        raise


@make_task(
    name="generate-curl",
    description="🔄 Generate curl command",
    group=http_group,
    alias="curl",
    input=[
        OptionInput(
            name="method",
            default="GET",
            options=["GET", "POST", "PUT", "DELETE", "PATCH"],
            prompt="HTTP method",
        ),
        StrInput(name="url", description="Target URL", prompt="Enter request URL"),
        StrInput(
            name="headers",
            description="Request headers (JSON)",
            prompt="Enter headers as JSON",
            default="{}",
        ),
        StrInput(
            name="body",
            description="Request body (JSON)",
            prompt="Enter body as JSON",
            default="{}",
        ),
        OptionInput(
            name="verify_ssl",
            default="true",
            options=["true", "false"],
            description="Verify SSL certificate",
        ),
    ],
)
def generate_curl(ctx: AnyContext) -> str:
    import json
    import shlex

    try:
        # Prepare curl command parts
        parts = ["curl"]

        # Add method
        parts.extend(["-X", ctx.input.method])

        # Add headers
        if ctx.input.headers != "{}":
            headers = json.loads(ctx.input.headers)
            for key, value in headers.items():
                parts.extend(["-H", f"{key}: {value}"])

        # Add body
        if ctx.input.body != "{}":
            parts.extend(["--data-raw", shlex.quote(ctx.input.body)])

        # Add SSL verification
        if ctx.input.verify_ssl.lower() == "false":
            parts.append("--insecure")

        # Add URL
        parts.append(shlex.quote(ctx.input.url))

        # Join parts into command string
        curl_command = " ".join(parts)
        ctx.print(f"🔄 Curl command: {curl_command}")
        return curl_command
    except Exception as e:
        ctx.print_err(f"Failed to generate curl command: {e}")
        raise



from typing import Any, Dict

from zrb.builtin.group import jwt_group
from zrb.context.any_context import AnyContext
from zrb.input.option_input import OptionInput
from zrb.input.str_input import StrInput
from zrb.task.make_task import make_task


@make_task(
    name="encode-jwt",
    description="🔒 Encode JWT token",
    group=jwt_group,
    alias="encode",
    input=[
        StrInput(name="secret", prompt="Secret", default=""),
        StrInput(name="payload", prompt="Payload (JSON format)", default="{}"),
        OptionInput(
            name="algorithm",
            prompt="Algorithm",
            default="HS256",
            options=["HS256", "HS384", "HS512"],
        ),
    ],
)
def encode_jwt(ctx: AnyContext) -> str:
    import jwt

    try:
        payload = eval(ctx.input.payload)
        token = jwt.encode(
            payload=payload, key=ctx.input.secret, algorithm=ctx.input.algorithm
        )
        ctx.print(token)
        return token
    except Exception as e:
        ctx.print_err(f"Failed to encode JWT: {e}")
        raise


@make_task(
    name="decode-jwt",
    description="🔓 Decode JWT token",
    group=jwt_group,
    alias="decode",
    input=[
        StrInput(name="token", prompt="Token", default=""),
        StrInput(name="secret", prompt="Secret", default=""),
        OptionInput(
            name="algorithm",
            prompt="Algorithm",
            default="HS256",
            options=["HS256", "HS384", "HS512"],
        ),
    ],
)
def decode_jwt(ctx: AnyContext) -> Dict[str, Any]:
    import jwt

    try:
        payload = jwt.decode(
            jwt=ctx.input.token, key=ctx.input.secret, algorithms=[ctx.input.algorithm]
        )
        ctx.print(payload)
        return payload
    except Exception as e:
        ctx.print_err(f"Failed to decode JWT: {e}")
        raise


@make_task(
    name="validate-jwt",
    description="✅ Validate JWT token",
    group=jwt_group,
    alias="validate",
    input=[
        StrInput(name="token", prompt="Token", default=""),
        StrInput(name="secret", prompt="Secret", default=""),
        OptionInput(
            name="algorithm",
            prompt="Algorithm",
            default="HS256",
            options=["HS256", "HS384", "HS512"],
        ),
    ],
)
def validate_jwt(ctx: AnyContext) -> bool:
    import jwt

    try:
        jwt.decode(
            jwt=ctx.input.token, key=ctx.input.secret, algorithms=[ctx.input.algorithm]
        )
        ctx.print("✅ Token is valid")
        return True
    except jwt.ExpiredSignatureError:
        ctx.print_err("❌ Token has expired")
        return False
    except jwt.InvalidTokenError:
        ctx.print_err("❌ Invalid token")
        return False
    except Exception as e:
        ctx.print_err(f"❌ Validation failed: {e}")
        return False



import asyncio
from typing import TYPE_CHECKING, Any

from zrb.context.any_context import AnyContext
from zrb.session.any_session import AnySession
from zrb.util.attr import get_bool_attr
from zrb.util.run import run_async
from zrb.xcom.xcom import Xcom

if TYPE_CHECKING:
    from zrb.task.base_task import BaseTask


async def execute_task_chain(task: "BaseTask", session: AnySession):
    """
    Executes the task and its downstream successors if conditions are met.
    """
    if session.is_terminated or not session.is_allowed_to_run(task):
        return
    result = await execute_task_action(task, session)
    # Get next tasks
    nexts = session.get_next_tasks(task)
    if session.is_terminated or len(nexts) == 0:
        return result
    # Run next tasks asynchronously
    next_coros = [run_async(next_task.exec_chain(session)) for next_task in nexts]
    # Wait for the next tasks to complete. The result of the current task is returned.
    await asyncio.gather(*next_coros)
    return result


async def execute_task_action(task: "BaseTask", session: AnySession):
    """
    Executes a single task's action, handling conditions and readiness checks.
    """
    ctx = task.get_ctx(session)
    if not session.is_allowed_to_run(task):
        # Task is not allowed to run, skip it for now.
        # This will be triggered later if dependencies are met.
        ctx.log_info("Not allowed to run")
        return
    if not check_execute_condition(task, session):
        # Skip the task based on its execute_condition
        ctx.log_info("Marked as skipped (condition false)")
        session.get_task_status(task).mark_as_skipped()
        return
    # Wait for task to be ready (handles action execution, readiness checks)
    return await run_async(execute_action_until_ready(task, session))


def check_execute_condition(task: "BaseTask", session: AnySession) -> bool:
    """
    Evaluates the task's execute_condition attribute.
    """
    ctx = task.get_ctx(session)
    execute_condition_attr = getattr(task, "_execute_condition", True)
    return get_bool_attr(ctx, execute_condition_attr, True, auto_render=True)


async def execute_action_until_ready(task: "BaseTask", session: AnySession):
    """
    Manages the execution of the task's action, coordinating with readiness checks.
    """
    ctx = task.get_ctx(session)
    readiness_checks = task.readiness_checks
    readiness_check_delay = getattr(task, "_readiness_check_delay", 0.5)
    monitor_readiness = getattr(task, "_monitor_readiness", False)

    if not readiness_checks:  # Simplified check for empty list
        ctx.log_info("No readiness checks")
        # Task has no readiness check, execute action directly
        result = await run_async(execute_action_with_retry(task, session))
        # Mark ready only if the action completed successfully (not failed/cancelled)
        if session.get_task_status(task).is_completed:
            ctx.log_info("Marked as ready")
            session.get_task_status(task).mark_as_ready()
        return result

    # Start the task action and readiness checks concurrently
    ctx.log_info("Starting action and readiness checks")
    action_coro = asyncio.create_task(
        run_async(execute_action_with_retry(task, session))
    )

    await asyncio.sleep(readiness_check_delay)

    readiness_check_coros = [
        run_async(check.exec_chain(session)) for check in readiness_checks
    ]

    # Wait primarily for readiness checks to complete
    ctx.log_info("Waiting for readiness checks")
    readiness_passed = False
    try:
        # Gather results, but primarily interested in completion/errors
        await asyncio.gather(*readiness_check_coros)
        # Check if all readiness tasks actually completed successfully
        all_readiness_completed = all(
            session.get_task_status(check).is_completed for check in readiness_checks
        )
        if all_readiness_completed:
            ctx.log_info("Readiness checks completed successfully")
            readiness_passed = True
            # Mark task as ready only if checks passed and action didn't fail during checks
            if not session.get_task_status(task).is_failed:
                ctx.log_info("Marked as ready")
                session.get_task_status(task).mark_as_ready()
        else:
            ctx.log_warning(
                "One or more readiness checks did not complete successfully."
            )

    except Exception as e:
        ctx.log_error(f"Readiness check failed with exception: {e}")
        # If readiness checks fail with an exception, the task is not ready.
        # The action_coro might still be running or have failed.
        # execute_action_with_retry handles marking the main task status.

    # Defer the main action coroutine; it will be awaited later if needed
    session.defer_action(task, action_coro)

    # Start monitoring only if readiness passed and monitoring is enabled
    if readiness_passed and monitor_readiness:
        # Import dynamically to avoid circular dependency if monitoring imports execution
        from zrb.task.base.monitoring import monitor_task_readiness

        monitor_coro = asyncio.create_task(
            run_async(monitor_task_readiness(task, session, action_coro))
        )
        session.defer_monitoring(task, monitor_coro)

    # The result here is primarily about readiness check completion.
    # The actual task result is handled by the deferred action_coro.
    return None


async def execute_action_with_retry(task: "BaseTask", session: AnySession) -> Any:
    """
    Executes the task's core action (`_exec_action`) with retry logic,
    handling success (triggering successors) and failure (triggering fallbacks).
    """
    ctx = task.get_ctx(session)
    retries = getattr(task, "_retries", 2)
    retry_period = getattr(task, "_retry_period", 0)
    max_attempt = retries + 1
    ctx.set_max_attempt(max_attempt)

    for attempt in range(max_attempt):
        ctx.set_attempt(attempt + 1)
        if attempt > 0:
            ctx.log_info(f"Retrying in {retry_period}s...")
            await asyncio.sleep(retry_period)

        try:
            ctx.log_info("Marked as started")
            session.get_task_status(task).mark_as_started()

            # Execute the underlying action (which might be overridden in subclasses)
            # We call the task's _exec_action method directly here.
            result = await run_async(task._exec_action(ctx))

            ctx.log_info("Marked as completed")
            session.get_task_status(task).mark_as_completed()

            # Store result in XCom
            task_xcom: Xcom = ctx.xcom.get(task.name)
            task_xcom.push(result)

            # Skip fallbacks and execute successors on success
            skip_fallbacks(task, session)
            await run_async(execute_successors(task, session))
            return result

        except (asyncio.CancelledError, KeyboardInterrupt):
            ctx.log_warning("Task cancelled or interrupted")
            session.get_task_status(task).mark_as_failed()  # Mark as failed on cancel
            # Do not trigger fallbacks/successors on cancellation
            return  # Or re-raise? Depends on desired cancellation behavior

        except BaseException as e:
            ctx.log_error(f"Attempt {attempt + 1}/{max_attempt} failed: {e}")
            session.get_task_status(
                task
            ).mark_as_failed()  # Mark failed for this attempt

            if attempt < max_attempt - 1:
                # More retries available
                continue
            else:
                # Final attempt failed
                ctx.log_error("Marked as permanently failed")
                session.get_task_status(task).mark_as_permanently_failed()
                # Skip successors and execute fallbacks on permanent failure
                skip_successors(task, session)
                await run_async(execute_fallbacks(task, session))
                raise e  # Re-raise the exception after handling fallbacks


async def run_default_action(task: "BaseTask", ctx: AnyContext) -> Any:
    """
    Executes the specific action defined by the '_action' attribute for BaseTask.
    This is the default implementation called by BaseTask._exec_action.
    Subclasses like LLMTask override _exec_action with their own logic.
    """
    action = getattr(task, "_action", None)
    if action is None:
        ctx.log_debug("No action defined for this task.")
        return None
    if isinstance(action, str):
        # Render f-string action
        rendered_action = ctx.render(action)
        ctx.log_debug(f"Rendered action string: {rendered_action}")
        # Assuming string actions are meant to be returned as is.
        # If they need execution (e.g., shell commands), that logic would go here.
        return rendered_action
    elif callable(action):
        # Execute callable action
        ctx.log_debug(f"Executing callable action: {action.__name__}")
        return await run_async(action(ctx))
    else:
        ctx.log_warning(f"Unsupported action type: {type(action)}")
        return None


async def execute_successors(task: "BaseTask", session: "AnySession"):
    """Executes all successor tasks."""
    ctx = task.get_ctx(session)
    successors = task.successors
    if successors:
        ctx.log_info(f"Executing {len(successors)} successor(s)")
        successor_coros = [
            run_async(successor.exec_chain(session)) for successor in successors
        ]
        await asyncio.gather(*successor_coros)
    else:
        ctx.log_debug("No successors to execute.")


def skip_successors(task: "BaseTask", session: AnySession):
    """Marks all successor tasks as skipped."""
    ctx = task.get_ctx(session)
    successors = task.successors
    if successors:
        ctx.log_info(f"Skipping {len(successors)} successor(s)")
        for successor in successors:
            # Check if already skipped to avoid redundant logging/state changes
            if not session.get_task_status(successor).is_skipped:
                session.get_task_status(successor).mark_as_skipped()


async def execute_fallbacks(task: "BaseTask", session: AnySession):
    """Executes all fallback tasks."""
    ctx = task.get_ctx(session)
    fallbacks = task.fallbacks
    if fallbacks:
        ctx.log_info(f"Executing {len(fallbacks)} fallback(s)")
        fallback_coros = [
            run_async(fallback.exec_chain(session)) for fallback in fallbacks
        ]
        await asyncio.gather(*fallback_coros)
    else:
        ctx.log_debug("No fallbacks to execute.")


def skip_fallbacks(task: "BaseTask", session: AnySession):
    """Marks all fallback tasks as skipped."""
    ctx = task.get_ctx(session)
    fallbacks = task.fallbacks
    if fallbacks:
        ctx.log_info(f"Skipping {len(fallbacks)} fallback(s)")
        for fallback in fallbacks:
            # Check if already skipped
            if not session.get_task_status(fallback).is_skipped:
                session.get_task_status(fallback).mark_as_skipped()



import re

NON_ALPHA_NUM = re.compile(r"[^a-zA-Z0-9]+")
TRUE_STRS = ["true", "1", "yes", "y", "active", "on"]
FALSE_STRS = ["false", "0", "no", "n", "inactive", "off"]


def double_quote(input_string: str) -> str:
    """
    Double quote a string, escaping any existing double quotes or backslashes.

    Args:
        input_string (str): The string to double quote.

    Returns:
        str: The double-quoted string with necessary characters escaped.
    """
    # Escape necessary characters: backslashes and double quotes
    escaped_string = re.sub(r'([\\"])', r"\\\1", input_string)
    # Wrap in double quotes
    return f'"{escaped_string}"'


def to_boolean(text: str) -> bool:
    """
    Convert a string representation to a boolean value.

    Recognizes common true/false strings (case-insensitive).

    Args:
        text (str): The string to convert.

    Returns:
        bool: The boolean representation of the string.

    Raises:
        Exception: If the string cannot be inferred as a boolean value.
    """
    if text.lower() in TRUE_STRS:
        return True
    if text.lower() in FALSE_STRS:
        return False
    raise Exception(f'Cannot infer boolean value from "{text}"')


def to_camel_case(text: str | None) -> str:
    """
    Convert a string to camelCase.

    Args:
        text (str | None): The input string.

    Returns:
        str: The string in camelCase.
    """
    text = str(text) if text is not None else ""
    pascal = to_pascal_case(text)
    if len(pascal) == 0:
        return pascal
    return pascal[0].lower() + pascal[1:]


def to_pascal_case(text: str | None) -> str:
    """
    Convert a string to PascalCase.

    Args:
        text (str | None): The input string.

    Returns:
        str: The string in PascalCase.
    """
    text = str(text) if text is not None else ""
    text = _to_alphanum(text)
    return "".join(
        [x.lower().capitalize() for x in _to_space_separated(text).split(" ")]
    )


def to_kebab_case(text: str | None) -> str:
    """
    Convert a string to kebab-case.

    Args:
        text (str | None): The input string.

    Returns:
        str: The string in kebab-case.
    """
    text = str(text) if text is not None else ""
    text = _to_alphanum(text)
    return "-".join([x.lower() for x in _to_space_separated(text).split(" ")])


def to_snake_case(text: str | None) -> str:
    """
    Convert a string to snake_case.

    Args:
        text (str | None): The input string.

    Returns:
        str: The string in snake_case.
    """
    text = str(text) if text is not None else ""
    text = _to_alphanum(text)
    return "_".join([x.lower() for x in _to_space_separated(text).split(" ")])


def _to_alphanum(text: str | None) -> str:
    """
    Convert a string to alphanumeric characters and spaces.

    Args:
        text (str | None): The input string.

    Returns:
        str: The string with non-alphanumeric characters replaced by spaces.
    """
    text = str(text) if text is not None else ""
    return NON_ALPHA_NUM.sub(" ", text)


def to_human_case(text: str | None) -> str:
    """
    Convert a string to human-readable case (space-separated words with capitalization).

    Args:
        text (str | None): The input string.

    Returns:
        str: The string in human-readable case.
    """
    text = str(text) if text is not None else ""
    return " ".join(
        [
            x.lower() if x.upper() != x else x
            for x in _to_space_separated(text).split(" ")
        ]
    )


def pluralize(noun: str) -> str:
    """
    Pluralize a given noun.

    Args:
        noun (str): The singular noun.

    Returns:
        str: The plural form of the noun.
    """
    # Irregular plural forms
    irregulars = {
        "foot": "feet",
        "tooth": "teeth",
        "child": "children",
        "person": "people",
        "man": "men",
        "woman": "women",
        "mouse": "mice",
        "goose": "geese",
        "ox": "oxen",
        "cactus": "cacti",
        "focus": "foci",
        "fungus": "fungi",
        "nucleus": "nuclei",
        "syllabus": "syllabi",
        "analysis": "analyses",
        "diagnosis": "diagnoses",
        "thesis": "theses",
        "crisis": "crises",
        "phenomenon": "phenomena",
        "criterion": "criteria",
    }

    # Handle irregular nouns
    if noun.lower() in irregulars:
        return irregulars[noun.lower()]
    # Handle words ending in 'y' preceded by a consonant
    if noun.endswith("y") and not re.match(r"[aeiou]y$", noun):
        return re.sub(r"y$", "ies", noun)
    # Handle words ending in 's', 'x', 'z', 'ch', or 'sh'
    if re.search(r"(s|x|z|ch|sh)$", noun):
        return noun + "es"
    # Handle words ending in 'f' or 'fe'
    if re.search(r"f$|fe$", noun):
        if noun.endswith("fe"):
            return re.sub(r"fe$", "ves", noun)
        return re.sub(r"f$", "ves", noun)
    # Handle general case for regular plurals
    return noun + "s"


def _to_space_separated(text: str | None) -> str:
    """
    Convert a string with various separators (-, _) and camel/pascal case to space-separated words.

    Args:
        text (str | None): The input string.

    Returns:
        str: The space-separated string.
    """
    text = str(text) if text is not None else ""
    text = text.replace("-", " ").replace("_", " ")
    parts = text.split(" ")
    new_parts = []
    for part in parts:
        new_part = ""
        for char_index, char in enumerate(part):
            is_first = char_index == 0
            is_last = char_index == len(part) - 1
            previous_char = part[char_index - 1] if not is_first else ""
            next_char = part[char_index + 1] if not is_last else ""
            if (
                char.isupper()
                and char != " "
                and (
                    (not is_last and next_char.islower())
                    or (not is_first and previous_char.islower())
                )
            ):
                new_part += " " + char
                continue
            new_part += char
        new_part = new_part.strip(" ")
        if new_part != "":
            new_parts.append(new_part)
    return " ".join(new_parts).strip(" ")



from typing import TYPE_CHECKING

from zrb.runner.web_config.config import WebConfig
from zrb.runner.web_schema.user import User

if TYPE_CHECKING:
    # Import Request only for type checking to reduce runtime dependencies
    from fastapi import Request


def get_user_by_credentials(
    web_config: WebConfig, username: str, password: str
) -> User | None:
    user = web_config.find_user_by_username(username)
    if user is None or not user.is_password_match(password):
        return None
    return user


async def get_user_from_request(
    web_config: WebConfig, request: "Request"
) -> User | None:
    from fastapi.security import OAuth2PasswordBearer

    if not web_config.enable_auth:
        return web_config.default_user
    # Normally we use "Depends"
    get_bearer_token = OAuth2PasswordBearer(tokenUrl="/api/v1/login", auto_error=False)
    bearer_token = await get_bearer_token(request)
    token_user = _get_user_from_token(web_config, bearer_token)
    if token_user is not None:
        return token_user
    cookie_user = _get_user_from_cookie(web_config, request)
    if cookie_user is not None:
        return cookie_user
    return web_config.default_user


def _get_user_from_cookie(web_config: WebConfig, request: "Request") -> User | None:
    token = request.cookies.get(web_config.access_token_cookie_name)
    if token:
        return _get_user_from_token(web_config, token)
    return None


def _get_user_from_token(web_config: WebConfig, token: str) -> User | None:
    try:
        from jose import jwt

        payload = jwt.decode(
            token,
            web_config.secret_key,
            options={"require_sub": True, "require_exp": True},
        )
        username: str = payload.get("sub")
        if username is None:
            return None
        user = web_config.find_user_by_username(username)
        if user is None:
            return None
        return user
    except Exception:
        return None



import re
from typing import Any


def fstring_format(template: str, data: dict[str, Any]) -> str:
    """
    Format a string template using f-string-like syntax with data from a dictionary.

    Expressions within curly braces `{}` are evaluated using the provided data.

    Args:
        template (str): The string template to format.
        data (dict[str, Any]): The dictionary containing data for expression evaluation.

    Returns:
        str: The formatted string.

    Raises:
        ValueError: If an expression in the template fails to evaluate or the
            template is invalid.
    """

    def replace_expr(match):
        """
        Helper function to evaluate a single expression found in the template.
        """
        expr = match.group(1)
        try:
            result = eval(expr, {}, data)
            return str(result)
        except Exception as e:
            raise ValueError(f"Failed to evaluate expression: {expr}: {e}")

    # Use regex to find and replace all expressions in curly braces
    pattern = r"\{([^}]+)\}"
    try:
        return re.sub(pattern, replace_expr, template)
    except Exception as e:
        raise ValueError(f"Failed to parse template: {template}: {e}")



import os
from typing import TYPE_CHECKING

from zrb.config import CFG
from zrb.group.any_group import AnyGroup
from zrb.runner.web_config.config import WebConfig
from zrb.runner.web_util.html import (
    get_html_auth_link,
    get_html_subgroup_info,
    get_html_subtask_info,
)
from zrb.runner.web_util.user import get_user_from_request
from zrb.util.file import read_file
from zrb.util.string.format import fstring_format

if TYPE_CHECKING:
    # We want fastapi to only be loaded when necessary to decrease footprint
    from fastapi import FastAPI


def serve_home_page(
    app: "FastAPI",
    root_group: AnyGroup,
    web_config: WebConfig,
) -> None:
    from fastapi import Request
    from fastapi.responses import HTMLResponse

    # Serve homepage
    @app.get("/", response_class=HTMLResponse, include_in_schema=False)
    @app.get("/ui", response_class=HTMLResponse, include_in_schema=False)
    @app.get("/ui/", response_class=HTMLResponse, include_in_schema=False)
    async def home_page_ui(request: Request) -> HTMLResponse:
        _DIR = os.path.dirname(__file__)
        _GLOBAL_TEMPLATE = read_file(
            os.path.join(os.path.dirname(_DIR), "static", "global_template.html")
        )
        _VIEW_TEMPLATE = read_file(os.path.join(_DIR, "view.html"))
        web_title = CFG.WEB_TITLE if CFG.WEB_TITLE.strip() != "" else root_group.name
        web_jargon = (
            CFG.WEB_JARGON if CFG.WEB_JARGON.strip() != "" else root_group.description
        )
        user = await get_user_from_request(web_config, request)
        group_info = get_html_subgroup_info(user, "/ui/", root_group)
        task_info = get_html_subtask_info(user, "/ui/", root_group)
        auth_link = get_html_auth_link(user)
        return HTMLResponse(
            fstring_format(
                _GLOBAL_TEMPLATE,
                {
                    "web_title": web_title,
                    "content": fstring_format(
                        _VIEW_TEMPLATE,
                        {
                            "web_title": web_title,
                            "web_jargon": web_jargon,
                            "web_homepage_intro": CFG.WEB_HOMEPAGE_INTRO,
                            "group_info": group_info,
                            "task_info": task_info,
                            "name": root_group.name,
                            "description": root_group.description,
                            "auth_link": auth_link,
                        },
                    ),
                },
            )
        )



import re
from typing import Callable

from sqlalchemy import asc, desc
from sqlalchemy.sql import ClauseElement, ColumnElement
from sqlmodel import SQLModel


def create_default_sort_param_parser() -> (
    Callable[[SQLModel, str], list[ColumnElement]]
):
    def parse_sort_param(model: SQLModel, query: str) -> list[ColumnElement]:
        """
        Parse the sort parameter and return a list of SQLAlchemy ColumnElement objects.

        Args:
            model (Type[SQLModel]): The SQLModel class to be queried.
            query (str): A comma-separated string of fields to sort by.
                        Prefix a field with '-' for descending order.

        Returns:
            list[ColumnElement]: A list of SQLAlchemy ColumnElement objects representing the sort order.

        Example:
            parse_sort_param(UserModel, "name,-age")
            # Returns [asc(UserModel.name), desc(UserModel.age)]
        """
        sorts: list[ColumnElement] = []
        sort_parts = query.split(",")
        for part in sort_parts:
            if part.startswith("-"):
                key = part[1:]
                order = desc
            else:
                key = part
                order = asc
            if hasattr(model, key):
                sorts.append(order(getattr(model, key)))
        return sorts

    return parse_sort_param


def create_default_filter_param_parser() -> (
    Callable[[SQLModel, str], list[ClauseElement]]
):
    def parse_filter_param(model: SQLModel, query: str) -> list[ClauseElement]:
        r"""
        Parse the filter parameter and return a list of SQLAlchemy ClauseElement objects.

        Args:
            model (Type[SQLModel]): The SQLModel class to be queried.
            query (str): A comma-separated string of filters.
                        Each filter should be in the format "field:operator:value".
                        Use '\,' to escape commas within values.

        Returns:
            list[ClauseElement]: A list of SQLAlchemy ClauseElement objects representing the filters.

        Supported operators:
            eq: Equal to
            ne: Not equal to
            gt: Greater than
            gte: Greater than or equal to
            lt: Less than
            lte: Less than or equal to
            like: SQL LIKE (use % for wildcards)
            in: In a list of values (semicolon-separated)

        Example:
            parse_filter_param(UserModel, "age:gte:18,name:like:John%,role:in:admin;user,address:eq:123\, Main St.")
            # Returns [UserModel.age >= 18, UserModel.name.like("John%"), UserModel.role.in_(["admin", "user"]), UserModel.address == "123, Main St."]
        """
        filters: list[ClauseElement] = []
        filter_parts = split_unescaped(query)
        for part in filter_parts:
            match = re.match(r"(.+):(.+):(.+)", part)
            if match:
                key, op, value = match.groups()
                value = value.replace(r"\,", ",")  # Unescape commas in the value
                if hasattr(model, key):
                    column = getattr(model, key)
                    if op == "eq":
                        filters.append(column == value)
                    elif op == "ne":
                        filters.append(column != value)
                    elif op == "gt":
                        filters.append(column > value)
                    elif op == "gte":
                        filters.append(column >= value)
                    elif op == "lt":
                        filters.append(column < value)
                    elif op == "lte":
                        filters.append(column <= value)
                    elif op == "like":
                        filters.append(column.like(value))
                    elif op == "in":
                        filters.append(column.in_(value.split(";")))
        return filters

    return parse_filter_param


def split_unescaped(s: str, delimiter: str = ",") -> list[str]:
    return re.split(r"(?<!\\)" + re.escape(delimiter), s)



import asyncio
from typing import Any

from zrb.context.shared_context import SharedContext
from zrb.session.any_session import AnySession
from zrb.session.session import Session
from zrb.task.any_task import AnyTask
from zrb.task.base.context import fill_shared_context_envs, fill_shared_context_inputs
from zrb.util.run import run_async


async def run_and_cleanup(
    task: AnyTask,
    session: AnySession | None = None,
    str_kwargs: dict[str, str] = {},
) -> Any:
    """
    Wrapper for async_run that ensures session termination and cleanup of
    other concurrent asyncio tasks. This is the main entry point for `task.run()`.
    """
    # Ensure a session exists
    if session is None:
        session = Session(shared_ctx=SharedContext())

    # Create the main task execution coroutine
    main_task_coro = asyncio.create_task(run_task_async(task, session, str_kwargs))

    try:
        result = await main_task_coro
        return result
    except (asyncio.CancelledError, KeyboardInterrupt) as e:
        ctx = task.get_ctx(session)  # Get context for logging
        ctx.log_warning(f"Run cancelled/interrupted: {e}")
        raise  # Re-raise to propagate
    finally:
        # Ensure session termination if it exists and wasn't terminated by the run
        if session and not session.is_terminated:
            ctx = task.get_ctx(session)  # Get context for logging
            ctx.log_info("Terminating session after run completion/error.")
            session.terminate()

        # Clean up other potentially running asyncio tasks (excluding the main one)
        # Be cautious with blanket cancellation if other background tasks are expected
        pending = [
            t for t in asyncio.all_tasks() if t is not main_task_coro and not t.done()
        ]
        if pending:
            ctx = task.get_ctx(session)  # Get context for logging
            ctx.log_debug(f"Cleaning up {len(pending)} pending asyncio tasks...")
            for t in pending:
                t.cancel()
            try:
                # Give cancelled tasks a moment to process cancellation
                await asyncio.wait(pending, timeout=1.0)
            except asyncio.CancelledError:
                # Expected if tasks handle cancellation promptly
                pass
            except Exception as cleanup_exc:
                # Log errors during cleanup if necessary
                ctx.log_warning(f"Error during task cleanup: {cleanup_exc}")


async def run_task_async(
    task: AnyTask,
    session: AnySession | None = None,
    str_kwargs: dict[str, str] = {},
) -> Any:
    """
    Asynchronous entry point for running a task (`task.async_run()`).
    Sets up the session and initiates the root task execution chain.
    """
    if session is None:
        session = Session(shared_ctx=SharedContext())

    # Populate shared context with inputs and environment variables
    fill_shared_context_inputs(task, session.shared_ctx, str_kwargs)
    fill_shared_context_envs(session.shared_ctx)  # Inject OS env vars

    # Start the execution chain from the root tasks
    result = await task.exec_root_tasks(session)
    return result


async def execute_root_tasks(task: AnyTask, session: AnySession):
    """
    Identifies and executes the root tasks required for the main task,
    manages session state logging, and handles overall execution flow.
    """
    session.set_main_task(task)
    session.state_logger.write(session.as_state_log())  # Initial state log
    ctx = task.get_ctx(session)  # Get context early for logging

    log_state_task = None
    try:
        # Start background state logging
        log_state_task = asyncio.create_task(log_session_state(task, session))

        # Identify root tasks allowed to run
        root_tasks = [
            t for t in session.get_root_tasks(task) if session.is_allowed_to_run(t)
        ]

        if not root_tasks:
            ctx.log_info("No root tasks to execute for this task.")
            # If the main task itself should run even with no explicit roots?
            # Current logic seems to imply if no roots, nothing runs.
            session.terminate()  # Terminate if nothing to run
            return None

        ctx.log_info(f"Executing {len(root_tasks)} root task(s)")
        root_task_coros = [
            # Assuming exec_chain exists on AnyTask (it's abstract)
            run_async(root_task.exec_chain(session))
            for root_task in root_tasks
        ]

        # Wait for all root chains to complete
        await asyncio.gather(*root_task_coros)

        # Wait for any deferred actions (like long-running task bodies)
        ctx.log_info("Waiting for deferred actions...")
        await session.wait_deferred()
        ctx.log_info("Deferred actions complete.")

        # Final termination and logging
        session.terminate()
        if log_state_task and not log_state_task.done():
            await log_state_task  # Ensure final state is logged
        ctx.log_info("Session finished.")
        return session.final_result

    except IndexError:
        # This might occur if get_root_tasks fails unexpectedly
        ctx.log_error(
            "IndexError during root task execution, potentially session issue."
        )
        session.terminate()  # Ensure termination on error
        return None
    except (asyncio.CancelledError, KeyboardInterrupt):
        ctx.log_warning("Session execution cancelled or interrupted.")
        # Session termination happens in finally block
        return None  # Indicate abnormal termination
    finally:
        # Ensure termination and final state logging regardless of outcome
        if not session.is_terminated:
            session.terminate()
        # Ensure the state logger task is awaited/cancelled properly
        if log_state_task:
            if not log_state_task.done():
                log_state_task.cancel()
                try:
                    await log_state_task
                except asyncio.CancelledError:
                    pass  # Expected cancellation
            # Log final state after ensuring logger task is finished
            session.state_logger.write(session.as_state_log())
        else:
            # Log final state even if logger task didn't start
            session.state_logger.write(session.as_state_log())

        ctx.log_debug(f"Final session state: {session}")  # Log final session details


async def log_session_state(task: AnyTask, session: AnySession):
    """
    Periodically logs the session state until the session is terminated.
    """
    try:
        while not session.is_terminated:
            session.state_logger.write(session.as_state_log())
            await asyncio.sleep(0.1)  # Log interval
        # Log one final time after termination signal
        session.state_logger.write(session.as_state_log())
    except asyncio.CancelledError:
        # Log final state on cancellation too
        session.state_logger.write(session.as_state_log())
        ctx = task.get_ctx(session)
        ctx.log_debug("Session state logger cancelled.")
    except Exception as e:
        # Log any unexpected errors in the logger itself
        ctx = task.get_ctx(session)
        ctx.log_error(f"Error in session state logger: {e}")



from zrb.builtin.group import llm_group
from zrb.builtin.llm.chat_session import get_llm_ask_input_mapping, read_user_prompt
from zrb.builtin.llm.history import read_chat_conversation, write_chat_conversation
from zrb.builtin.llm.input import PreviousSessionInput
from zrb.builtin.llm.tool.api import get_current_location, get_current_weather
from zrb.builtin.llm.tool.cli import run_shell_command
from zrb.builtin.llm.tool.file import (
    apply_diff,
    list_files,
    read_from_file,
    search_files,
    write_to_file,
)
from zrb.builtin.llm.tool.sub_agent import create_sub_agent_tool
from zrb.builtin.llm.tool.web import (
    create_search_internet_tool,
    open_web_page,
    search_arxiv,
    search_wikipedia,
)
from zrb.callback.callback import Callback
from zrb.config import CFG
from zrb.input.bool_input import BoolInput
from zrb.input.str_input import StrInput
from zrb.input.text_input import TextInput
from zrb.task.base_trigger import BaseTrigger
from zrb.task.llm_task import LLMTask

_llm_ask_inputs = [
    StrInput(
        "model",
        description="LLM Model",
        prompt="LLM Model",
        default="",
        allow_positional_parsing=False,
        always_prompt=False,
        allow_empty=True,
    ),
    StrInput(
        "base-url",
        description="LLM API Base URL",
        prompt="LLM API Base URL",
        default="",
        allow_positional_parsing=False,
        always_prompt=False,
        allow_empty=True,
    ),
    StrInput(
        "api-key",
        description="LLM API Key",
        prompt="LLM API Key",
        default="",
        allow_positional_parsing=False,
        always_prompt=False,
        allow_empty=True,
    ),
    TextInput(
        "system-prompt",
        description="System prompt",
        prompt="System prompt",
        default="",
        allow_positional_parsing=False,
        always_prompt=False,
    ),
    BoolInput(
        "start-new",
        description="Start new conversation (LLM will forget everything)",
        prompt="Start new conversation (LLM will forget everything)",
        default=False,
        allow_positional_parsing=False,
        always_prompt=False,
    ),
    TextInput("message", description="User message", prompt="Your message"),
    PreviousSessionInput(
        "previous-session",
        description="Previous conversation session",
        prompt="Previous conversation session (can be empty)",
        allow_positional_parsing=False,
        allow_empty=True,
        always_prompt=False,
    ),
]

llm_ask: LLMTask = llm_group.add_task(
    LLMTask(
        name="llm-ask",
        input=_llm_ask_inputs,
        description="❓ Ask LLM",
        model=lambda ctx: None if ctx.input.model.strip() == "" else ctx.input.model,
        model_base_url=lambda ctx: (
            None if ctx.input.base_url.strip() == "" else ctx.input.base_url
        ),
        model_api_key=lambda ctx: (
            None if ctx.input.api_key.strip() == "" else ctx.input.api_key
        ),
        conversation_history_reader=read_chat_conversation,
        conversation_history_writer=write_chat_conversation,
        system_prompt=lambda ctx: (
            None if ctx.input.system_prompt.strip() == "" else ctx.input.system_prompt
        ),
        message="{ctx.input.message}",
        retries=0,
    ),
    alias="ask",
)

llm_group.add_task(
    BaseTrigger(
        name="llm-chat",
        input=_llm_ask_inputs,
        description="💬 Chat with LLM",
        queue_name="ask_trigger",
        action=read_user_prompt,
        callback=Callback(
            task=llm_ask,
            input_mapping=get_llm_ask_input_mapping,
            result_queue="ask_result",
            error_queue="ask_error",
            session_name_queue="ask_session_name",
        ),
        retries=0,
        cli_only=True,
    ),
    alias="chat",
)

if CFG.LLM_ALLOW_ACCESS_LOCAL_FILE:
    llm_ask.append_tool(
        list_files,
        read_from_file,
        write_to_file,
        search_files,
        apply_diff,
        create_sub_agent_tool(
            tool_name="analyze_file",
            tool_description="\n".join(
                [
                    "Analyze file using LLM capability.",
                    "This tool can do:",
                    "- summarization",
                    "- outline/structure extraction",
                    "- code review",
                    "- other tasks requiring deep understanding.",
                    "Always use this tool to get deep understanding of file content",
                ]
            ),
            sub_agent_system_prompt="\n".join(
                [
                    "You are a file analyzer assistant",
                    "Your goal is to help the main asisstant by reading file",
                    "and perform necessary indepth analysis required by the main assistant",
                ]
            ),
            sub_agent_tools=[read_from_file, search_files],
        ),
    )

if CFG.LLM_ALLOW_ACCESS_SHELL:
    llm_ask.append_tool(run_shell_command)

if CFG.LLM_ALLOW_ACCESS_INTERNET:
    llm_ask.append_tool(open_web_page, search_wikipedia, search_arxiv)
    if CFG.SERP_API_KEY != "":
        llm_ask.append_tool(create_search_internet_tool(CFG.SERP_API_KEY))
    llm_ask.append_tool(get_current_location, get_current_weather)



from abc import ABC, abstractmethod
from collections.abc import Callable

from zrb.attr.type import StrAttr, fstring
from zrb.context.context import Context
from zrb.util.attr import get_str_attr
from zrb.util.file import read_file


class AnyCmdVal(ABC):
    @abstractmethod
    def to_str(self, ctx: Context) -> str:
        pass


class CmdPath(AnyCmdVal):
    def __init__(self, path: StrAttr, auto_render: bool = True):
        self._path = path
        self._auto_render = auto_render

    def to_str(self, ctx: Context) -> str:
        file_path = get_str_attr(ctx, self._path, "", self._auto_render)
        return read_file(file_path)


class Cmd(AnyCmdVal):
    def __init__(self, cmd: StrAttr, auto_render: bool = True):
        self._cmd = cmd
        self._auto_render = auto_render

    def to_str(self, ctx: Context) -> str:
        return get_str_attr(ctx, self._cmd, "", self._auto_render)


SingleCmdVal = AnyCmdVal | fstring | Callable[[Context], str]
CmdVal = SingleCmdVal | list[SingleCmdVal]



import os

from zrb.config import CFG
from zrb.group.any_group import AnyGroup
from zrb.runner.web_schema.user import User
from zrb.runner.web_util.html import get_html_auth_link
from zrb.util.file import read_file
from zrb.util.string.format import fstring_format


def show_error_page(user: User, root_group: AnyGroup, status_code: int, message: str):
    from fastapi.responses import HTMLResponse

    _DIR = os.path.dirname(__file__)
    _GLOBAL_TEMPLATE = read_file(
        os.path.join(os.path.dirname(_DIR), "static", "global_template.html")
    )
    _VIEW_TEMPLATE = read_file(os.path.join(_DIR, "view.html"))
    auth_link = get_html_auth_link(user)
    return HTMLResponse(
        fstring_format(
            _GLOBAL_TEMPLATE,
            {
                "web_title": CFG.WEB_TITLE,
                "content": fstring_format(
                    _VIEW_TEMPLATE,
                    {
                        "name": root_group.name,
                        "description": root_group.description,
                        "auth_link": auth_link,
                        "error_status_code": status_code,
                        "error_message": message,
                    },
                ),
            },
        ),
        status_code=status_code,
    )



from zrb.config import CFG
from zrb.runner.web_config.config import WebConfig

web_config = WebConfig(
    port=CFG.WEB_HTTP_PORT,
    secret_key=CFG.WEB_SECRET_KEY,
    access_token_expire_minutes=CFG.WEB_AUTH_ACCESS_TOKEN_EXPIRE_MINUTES,
    refresh_token_expire_minutes=CFG.WEB_AUTH_REFRESH_TOKEN_EXPIRE_MINUTES,
    access_token_cookie_name=CFG.WEB_ACCESS_TOKEN_COOKIE_NAME,
    refresh_token_cookie_name=CFG.WEB_REFRESH_TOKEN_COOKIE_NAME,
    enable_auth=CFG.WEB_ENABLE_AUTH,
    super_admin_username=CFG.WEB_SUPER_ADMIN_USERNAME,
    super_admin_password=CFG.WEB_SUPER_ADMIN_PASSWORD,
    guest_username=CFG.WEB_GUEST_USERNAME,
)



from typing import Any

from pydantic import BaseModel


class TaskStatusHistoryStateLog(BaseModel):
    status: str
    time: str


class TaskStatusStateLog(BaseModel):
    history: list[TaskStatusHistoryStateLog]
    is_started: bool
    is_ready: bool
    is_completed: bool
    is_skipped: bool
    is_failed: bool
    is_permanently_failed: bool
    is_terminated: bool


class SessionStateLog(BaseModel):
    name: str
    start_time: str
    main_task_name: str
    path: list[str]
    input: dict[str, Any]
    final_result: str
    finished: bool
    log: list[str]
    task_status: dict[str, TaskStatusStateLog]


class SessionStateLogList(BaseModel):
    total: int
    data: list[SessionStateLog]



import datetime
import re
import shutil

from pydantic import BaseModel, Field, model_validator

from zrb.util.cli.style import (
    stylize_bold_yellow,
    stylize_cyan,
    stylize_faint,
    stylize_magenta,
    stylize_yellow,
)
from zrb.util.file import read_file, write_file
from zrb.util.string.name import get_random_name

_DATE_TIME_STR_WIDTH = 14
_MAX_DESCRIPTION_WIDTH = 70
_PRIORITY_WIDTH = 3
_COMPLETED_WIDTH = 3
_COMPLETED_AT_WIDTH = _DATE_TIME_STR_WIDTH
_CREATED_AT_WIDTH = _DATE_TIME_STR_WIDTH
_GAP_WIDTH = 2


class TodoTaskModel(BaseModel):
    priority: str | None = Field("D", pattern=r"^[A-Z]$")  # Priority like A, B, ...
    completed: bool = False  # True if completed, False otherwise
    description: str  # Main task description
    projects: list[str] = []  # List of projects (e.g., +Project)
    contexts: list[str] = []  # List of contexts (e.g., @Context)
    keyval: dict[str, str] = {}  # Special key (e.g., due:2016-05-30)
    creation_date: datetime.date | None = None  # Creation date
    completion_date: datetime.date | None = None  # Completion date

    @model_validator(mode="before")
    def validate_dates(cls, values):
        completion_date = values.get("completion_date")
        creation_date = values.get("creation_date")
        if completion_date and not creation_date:
            raise ValueError(
                "creation_date must be specified if completion_date is set."
            )
        return values

    def get_additional_info_length(self):
        """
        Calculate the length of the additional information string (projects, contexts, keyval).

        Returns:
            int: The length of the combined additional information string.
        """
        results = []
        for project in self.projects:
            results.append(f"@{project}")
        for context in self.contexts:
            results.append(f"+{context}")
        for key, val in self.keyval.items():
            results.append(f"{key}:{val}")
        return len(", ".join(results))


TODO_TXT_PATTERN = re.compile(
    r"^(?P<status>x)?\s*"  # Optional completion mark ('x')
    r"(?:\((?P<priority>[A-Z])\)\s+)?"  # Optional priority (e.g., '(A)')
    r"(?P<date1>\d{4}-\d{2}-\d{2})?\s*"  # Optional first date
    r"(?P<date2>\d{4}-\d{2}-\d{2})?\s*"  # Optional second date
    r"(?P<description>.*?)$"  # Main description
)


def cascade_todo_task(todo_task: TodoTaskModel):
    """
    Populate default values for a TodoTaskModel if they are missing.

    Args:
        todo_task (TodoTaskModel): The todo task model to cascade.

    Returns:
        TodoTaskModel: The todo task model with default values populated.
    """
    if todo_task.creation_date is None:
        todo_task.creation_date = datetime.date.today()
    if "id" not in todo_task.keyval:
        todo_task.keyval["id"] = get_random_name()
    return todo_task


def select_todo_task(
    todo_list: list[TodoTaskModel], keyword: str
) -> TodoTaskModel | None:
    """
    Select a todo task from a list based on a keyword matching ID or description.

    Args:
        todo_list (list[TodoTaskModel]): The list of todo tasks.
        keyword (str): The keyword to search for.

    Returns:
        TodoTaskModel | None: The matched todo task, or None if no match is found.
    """
    for todo_task in todo_list:
        id = todo_task.keyval.get("id", "")
        if keyword.lower().strip() == id.lower().strip():
            return todo_task
    for todo_task in todo_list:
        description = todo_task.description
        if keyword.lower().strip() == description.lower().strip():
            return todo_task
    for todo_task in todo_list:
        id = todo_task.keyval.get("id", "")
        if keyword.lower().strip() in id.lower().strip():
            return todo_task
    for todo_task in todo_list:
        description = todo_task.description
        if keyword.lower().strip() in description.lower().strip():
            return todo_task
    return None


def load_todo_list(todo_file_path: str) -> list[TodoTaskModel]:
    """
    Load a list of todo tasks from a todo.txt file.

    Args:
        todo_file_path (str): The path to the todo.txt file.

    Returns:
        list[TodoTaskModel]: A sorted list of todo tasks.
    """
    todo_lines = read_file(todo_file_path).strip().split("\n")
    todo_list: list[TodoTaskModel] = []
    for todo_line in todo_lines:
        todo_line = todo_line.strip()
        if todo_line == "":
            continue
        todo_task = line_to_todo_task(todo_line)
        todo_list.append(todo_task)
    todo_list.sort(
        key=lambda task: (
            task.completed,
            task.priority if task.priority else "Z",
            task.projects[0] if task.projects else "zzz",
            task.creation_date if task.creation_date else datetime.date.max,
        )
    )
    return todo_list


def save_todo_list(todo_file_path: str, todo_list: list[TodoTaskModel]):
    """
    Save a list of todo tasks to a todo.txt file.

    Args:
        todo_file_path (str): The path to the todo.txt file.
        todo_list (list[TodoTaskModel]): The list of todo tasks to save.
    """
    write_file(
        todo_file_path, [todo_task_to_line(todo_task) for todo_task in todo_list]
    )


def line_to_todo_task(line: str) -> TodoTaskModel:
    """Parses a single todo.txt line into a TodoTask model."""
    match = TODO_TXT_PATTERN.match(line)
    if not match:
        raise ValueError(f"Invalid todo.txt line: {line}")
    groups = match.groupdict()
    # Extract completion status
    is_completed = groups["status"] == "x"
    # Extract dates
    date1 = _parse_date(groups["date1"])
    date2 = _parse_date(groups["date2"])
    # Determine creation_date and completion_date
    completion_date, creation_date = None, None
    if date2 is None:
        creation_date = date1
    else:
        completion_date = date1
        creation_date = date2
    # Extract and clean description
    raw_description = groups["description"] or ""
    projects = re.findall(r"\+(\S+)", raw_description)
    contexts = re.findall(r"@(\S+)", raw_description)
    keyval = {}
    for keyval_str in re.findall(r"(\S+:\S+)", raw_description):
        key, val = keyval_str.split(":", 1)
        keyval[key] = val
    description = re.sub(r"\s*\+\S+|\s*@\S+|\s*\S+:\S+", "", raw_description).strip()
    return TodoTaskModel(
        priority=groups["priority"],
        completed=is_completed,
        description=description,
        projects=projects,
        contexts=contexts,
        keyval=keyval,
        creation_date=creation_date,
        completion_date=completion_date,
    )


def _parse_date(date_str: str | None) -> datetime.date | None:
    """
    Parses a date string in the format YYYY-MM-DD.

    Args:
        date_str (str | None): The date string to parse.

    Returns:
        datetime.date | None: The parsed date object, or None if the input is None.
    """
    if date_str:
        return datetime.date.fromisoformat(date_str)
    return None


def todo_task_to_line(task: TodoTaskModel) -> str:
    """
    Converts a TodoTask instance back into a todo.txt formatted line.

    Args:
        task (TodoTaskModel): The todo task instance.

    Returns:
        str: The todo.txt formatted line.
    """
    parts = []
    # Add completion mark if task is completed
    if task.completed:
        parts.append("x")
    # Add priority if present
    if task.priority:
        parts.append(f"({task.priority})")
    # Add completion and creation dates if present
    if task.completion_date:
        parts.append(task.completion_date.isoformat())
    if task.creation_date:
        parts.append(task.creation_date.isoformat())
    # Add description
    parts.append(task.description)
    # Append projects
    for project in task.projects:
        parts.append(f"+{project}")
    # Append contexts
    for context in task.contexts:
        parts.append(f"@{context}")
    # Append keyval
    for key, val in task.keyval.items():
        parts.append(f"{key}:{val}")
    # Join all parts with a space
    return " ".join(parts)


def get_visual_todo_list(todo_list: list[TodoTaskModel], filter: str) -> str:
    """
    Generate a visual representation of a filtered todo list.

    Args:
        todo_list (list[TodoTaskModel]): The list of todo tasks.
        filter (str): The filter string in todo.txt format.

    Returns:
        str: A formatted string representing the filtered todo list.
    """
    todo_filter = line_to_todo_task(filter)
    filtered_todo_list = []
    for todo_task in todo_list:
        filter_description = todo_filter.description.lower().strip()
        if (
            filter_description != ""
            and filter_description not in todo_task.description.lower()
        ):
            continue
        if not all(context in todo_task.contexts for context in todo_filter.contexts):
            continue
        if not all(project in todo_task.projects for project in todo_filter.projects):
            continue
        if not all(
            key in todo_task.keyval and todo_task.keyval[key] == val
            for key, val in todo_filter.keyval.items()
        ):
            continue
        filtered_todo_list.append(todo_task)
    if len(filtered_todo_list) == 0:
        return "\n".join(["", "  Empty todo list... 🌵🦖", ""])
    max_desc_length = max(
        len(todo_task.description) for todo_task in filtered_todo_list
    )
    if max_desc_length < len("DESCRIPTION"):
        max_desc_length = len("DESCRIPTION")
    if max_desc_length > _MAX_DESCRIPTION_WIDTH:
        max_desc_length = _MAX_DESCRIPTION_WIDTH
    max_additional_info_length = max(
        todo_task.get_additional_info_length() for todo_task in filtered_todo_list
    )
    if max_additional_info_length < len("PROJECT/CONTEXT/OTHERS"):
        max_additional_info_length = len("PROJECT/CONTEXT/OTHERS")
    terminal_width, _ = shutil.get_terminal_size()
    # Headers
    results = [
        stylize_faint(
            get_visual_todo_header(
                terminal_width, max_desc_length, max_additional_info_length
            )
        )
    ]
    for todo_task in filtered_todo_list:
        results.append(
            get_visual_todo_line(
                terminal_width, max_desc_length, max_additional_info_length, todo_task
            )
        )
    return "\n".join(results)


def get_visual_todo_header(
    terminal_width: int, max_desc_length: int, max_additional_info_length: int
) -> str:
    """
    Generate the header string for the visual todo list.

    Args:
        terminal_width (int): The width of the terminal.
        max_desc_length (int): The maximum length of the description column.
        max_additional_info_length (int): The maximum length of the additional info column.

    Returns:
        str: The formatted header string.
    """
    priority_caption = "".ljust(_PRIORITY_WIDTH)
    completed_caption = "".ljust(_COMPLETED_WIDTH)
    completed_at_caption = "COMPLETED AT".rjust(_COMPLETED_AT_WIDTH)
    created_at_caption = "CREATED_AT".ljust(_CREATED_AT_WIDTH)
    description_caption = "DESCRIPTION".ljust(
        min(max_desc_length, _MAX_DESCRIPTION_WIDTH)
    )
    additional_info_caption = "PROJECT/CONTEXT/OTHERS".ljust(max_additional_info_length)
    return _get_line_str(
        terminal_width=terminal_width,
        description_width=min(max_desc_length, _MAX_DESCRIPTION_WIDTH),
        additional_info_width=max_additional_info_length,
        priority=priority_caption,
        completed=completed_caption,
        completed_at=completed_at_caption,
        created_at=created_at_caption,
        description=description_caption,
        additional_info=additional_info_caption,
    )


def get_visual_todo_line(
    terminal_width: int,
    max_desc_length: int,
    max_additional_info_length: int,
    todo_task: TodoTaskModel,
) -> str:
    """
    Generate a single line string for a todo task in the visual todo list.

    Args:
        terminal_width (int): The width of the terminal.
        max_desc_length (int): The maximum length of the description column.
        max_additional_info_length (int): The maximum length of the additional info column.
        todo_task (TodoTaskModel): The todo task to format.

    Returns:
        str: The formatted line string for the todo task.
    """
    completed = "[x]" if todo_task.completed else "[ ]"
    priority = "   " if todo_task.priority is None else f"({todo_task.priority})"
    completed_at = stylize_yellow(_date_to_str(todo_task.completion_date))
    created_at = stylize_cyan(_date_to_str(todo_task.creation_date))
    description = todo_task.description
    if len(description) > max_desc_length:
        description = description[: max_desc_length - 4] + " ..."
    description = description.ljust(max_desc_length)
    description = description[:max_desc_length]
    if todo_task.completed:
        description = stylize_faint(description)
    elif "duration" in todo_task.keyval:
        description = stylize_bold_yellow(description)
    additional_info = ", ".join(
        [stylize_yellow(f"+{project}") for project in todo_task.projects]
        + [stylize_cyan(f"@{context}") for context in todo_task.contexts]
        + [stylize_magenta(f"{key}:{val}") for key, val in todo_task.keyval.items()]
    )
    return _get_line_str(
        terminal_width=terminal_width,
        description_width=min(max_desc_length, _MAX_DESCRIPTION_WIDTH),
        additional_info_width=max_additional_info_length,
        priority=priority,
        completed=completed,
        completed_at=completed_at,
        created_at=created_at,
        description=description,
        additional_info=additional_info,
    )
    if terminal_width <= 14 + max_desc_length + max_additional_info_length:
        return "  ".join([priority, completed, description])
    if terminal_width <= 36 + max_desc_length + max_additional_info_length:
        return "  ".join([priority, completed, description, additional_info])
    return "  ".join(
        [priority, completed, completed_at, created_at, description, additional_info]
    )


def _get_line_str(
    terminal_width: int,
    description_width: int,
    additional_info_width: int,
    priority: str,
    completed: str,
    completed_at: str,
    created_at: str,
    description: str,
    additional_info: str,
):
    """
    Helper function to format a line string based on terminal width.

    Args:
        terminal_width (int): The width of the terminal.
        description_width (int): The width of the description column.
        additional_info_width (int): The width of the additional info column.
        priority (str): The formatted priority string.
        completed (str): The formatted completed status string.
        completed_at (str): The formatted completed at date string.
        created_at (str): The formatted created at date string.
        description (str): The formatted description string.
        additional_info (str): The formatted additional info string.

    Returns:
        str: The formatted line string.
    """
    gap = "".ljust(_GAP_WIDTH)
    if terminal_width >= _get_minimum_width(
        [
            _PRIORITY_WIDTH,
            _COMPLETED_WIDTH,
            _COMPLETED_AT_WIDTH,
            _CREATED_AT_WIDTH,
            description_width,
            additional_info_width,
        ]
    ):
        return gap.join(
            [
                priority,
                completed,
                completed_at,
                created_at,
                description,
                additional_info,
            ]
        )
    if terminal_width >= _get_minimum_width(
        [_PRIORITY_WIDTH, _COMPLETED_WIDTH, description_width, additional_info_width]
    ):
        return gap.join([priority, completed, description, additional_info])
    if terminal_width >= _get_minimum_width(
        [
            _PRIORITY_WIDTH,
            _COMPLETED_WIDTH,
            _COMPLETED_AT_WIDTH,
            _CREATED_AT_WIDTH,
            description_width,
        ]
    ):
        return gap.join([priority, completed, completed_at, created_at, description])
    if terminal_width >= _get_minimum_width(
        [_PRIORITY_WIDTH, _COMPLETED_WIDTH, description_width]
    ):
        return gap.join([priority, completed, description])
    return gap.join([priority, description])


def _get_minimum_width(field_widths: list[int]) -> int:
    """
    Helper function to calculate the minimum width required for a list of fields.

    Args:
        field_widths (list[int]): A list of widths for each field.

    Returns:
        int: The minimum total width required.
    """
    gap_width = _GAP_WIDTH * (len(field_widths) - 1)
    return sum(field_width for field_width in field_widths) + gap_width


def get_visual_todo_card(
    todo_task: TodoTaskModel, log_work_list: list[dict[str, str]]
) -> str:
    """
    Generate a visual card representation of a todo task with log work.

    Args:
        todo_task (TodoTaskModel): The todo task to display.
        log_work_list (list[dict[str, str]]): A list of log work entries for the task.

    Returns:
        str: A formatted string representing the todo task card.
    """
    description = todo_task.description
    status = "TODO"
    if todo_task.completed:
        status = "DONE"
    elif "duration" in todo_task.keyval:
        status = "DOING"
    priority = todo_task.priority
    completed_at = (
        _date_to_str(todo_task.completion_date)
        if todo_task.completion_date is not None
        else ""
    )
    created_at = (
        _date_to_str(todo_task.creation_date)
        if todo_task.creation_date is not None
        else ""
    )
    log_work_str = "\n".join(
        [
            "  ".join(
                [
                    stylize_magenta(log_work.get("duration", "").strip().rjust(12)),
                    stylize_cyan(log_work.get("start", "").strip().rjust(20)),
                    log_work.get("log", "").strip(),
                ]
            )
            for log_work in log_work_list
        ]
    )
    detail = [
        f"{'📄 Description'.ljust(16)}: {description}",
        f"{'🎯 Priority'.ljust(16)}: {priority}",
        f"{'📊 Status'.ljust(16)}: {status}",
        f"{'📅 Created at'.ljust(16)}: {created_at}",
        f"{'✅ Completed at'.ljust(16)}: {completed_at}",
    ]
    if log_work_str != "":
        detail.append(
            stylize_faint("  ".join(["Time Spent".rjust(12), "Start".rjust(20), "Log"]))
        )
        detail.append(log_work_str)
    return "\n".join(detail)


def _date_to_str(date: datetime.date | None) -> str:
    """
    Helper function to format a date object as a string.

    Args:
        date (datetime.date | None): The date object to format.

    Returns:
        str: The formatted date string, or an empty string if the input is None.
    """
    if date is None:
        return "".ljust(14)
    return date.strftime("%a %Y-%m-%d")


def add_duration(duration1: str, duration2: str) -> str:
    """
    Add two duration strings.

    Args:
        duration1 (str): The first duration string.
        duration2 (str): The second duration string.

    Returns:
        str: The sum of the two durations as a formatted string.
    """
    total_seconds = parse_duration(duration1) + parse_duration(duration2)
    # Format and return the result
    return _format_duration(total_seconds)


def parse_duration(duration: str) -> int:
    """
    Parse a duration string into total seconds.

    Args:
        duration (str): The duration string to parse.

    Returns:
        int: The total duration in seconds.
    """
    units = {"M": 2592000, "w": 604800, "d": 86400, "h": 3600, "m": 60, "s": 1}
    total_seconds = 0
    match = re.findall(r"(\d+)([Mwdhms])", duration)
    for value, unit in match:
        total_seconds += int(value) * units[unit]
    return total_seconds


def _format_duration(total_seconds: int) -> str:
    """
    Format total seconds into a duration string.

    Args:
        total_seconds (int): The total duration in seconds.

    Returns:
        str: The formatted duration string.
    """
    units = [
        ("w", 604800),  # 7 days in a week
        ("d", 86400),  # 24 hours in a day
        ("h", 3600),  # 60 minutes in an hour
        ("m", 60),  # 60 seconds in a minute
        ("s", 1),  # seconds
    ]
    result = []
    for unit, value_in_seconds in units:
        if total_seconds >= value_in_seconds:
            amount, total_seconds = divmod(total_seconds, value_in_seconds)
            result.append(f"{amount}{unit}")
    return "".join(result) if result else "0s"



import os

from fastapi import FastAPI
from my_app_name.common.db_engine_factory import db_engine
from my_app_name.common.util.app import (
    create_default_app_lifespan,
    get_default_app_title,
    serve_docs,
    serve_static_dir,
)
from my_app_name.config import (
    APP_GATEWAY_FAVICON_PATH,
    APP_GATEWAY_TITLE,
    APP_GATEWAY_VIEW_PATH,
    APP_MODE,
    APP_MODULES,
    APP_VERSION,
)

app_title = get_default_app_title(APP_GATEWAY_TITLE, mode=APP_MODE, modules=APP_MODULES)
app = FastAPI(
    title=app_title,
    version=APP_VERSION,
    lifespan=create_default_app_lifespan(db_engine),
    docs_url=None,
)

serve_static_dir(app, os.path.join(APP_GATEWAY_VIEW_PATH, "static"))
serve_docs(app, app_title=app_title, favicon_url=APP_GATEWAY_FAVICON_PATH)



import sys
from typing import Any

from zrb.config import CFG
from zrb.context.any_context import AnyContext
from zrb.context.shared_context import SharedContext
from zrb.group.any_group import AnyGroup
from zrb.group.group import Group
from zrb.runner.common_util import get_run_kwargs
from zrb.runner.web_app import create_web_app
from zrb.runner.web_config.config_factory import web_config
from zrb.session.session import Session
from zrb.session_state_logger.session_state_logger_factory import session_state_logger
from zrb.task.any_task import AnyTask
from zrb.task.make_task import make_task
from zrb.util.cli.style import (
    stylize_bold_yellow,
    stylize_faint,
    stylize_section_header,
)
from zrb.util.group import extract_node_from_args, get_non_empty_subgroups, get_subtasks
from zrb.util.init_path import get_init_path_list
from zrb.util.string.conversion import double_quote


class Cli(Group):

    def __init__(self):
        super().__init__(name="_zrb")

    @property
    def name(self):
        return CFG.ROOT_GROUP_NAME

    @property
    def description(self):
        return CFG.ROOT_GROUP_DESCRIPTION

    @property
    def banner(self) -> str:
        return CFG.BANNER

    def run(self, args: list[str] = []):
        kwargs, args = self._extract_kwargs_from_args(args)
        node, node_path, args = extract_node_from_args(self, args)
        if isinstance(node, AnyGroup):
            self._show_group_info(node)
            return
        if "h" in kwargs or "help" in kwargs:
            self._show_task_info(node)
            return
        run_kwargs = get_run_kwargs(task=node, args=args, kwargs=kwargs, cli_mode=True)
        try:
            result = self._run_task(node, args, run_kwargs)
            if result is not None:
                print(result)
            return result
        finally:
            run_command = self._get_run_command(node_path, run_kwargs)
            self._print_run_command(run_command)

    def _print_run_command(self, run_command: str):
        print(
            stylize_faint("To run again:"),
            stylize_bold_yellow(run_command),
            file=sys.stderr,
        )

    def _get_run_command(self, node_path: list[str], run_kwargs: dict[str, str]) -> str:
        parts = [self.name] + node_path
        if len(run_kwargs) > 0:
            parts += [
                self._get_run_command_param(key, val) for key, val in run_kwargs.items()
            ]
        return " ".join(parts)

    def _get_run_command_param(self, key: str, val: str) -> str:
        if '"' in val or "'" in val or " " in val or val == "":
            return f"--{key} {double_quote(val)}"
        return f"--{key} {val}"

    def _run_task(
        self, task: AnyTask, args: list[str], run_kwargs: dict[str, str]
    ) -> tuple[Any]:
        shared_ctx = SharedContext(args=args)
        for task_input in task.inputs:
            if task_input.name in run_kwargs:
                task_input.update_shared_context(
                    shared_ctx, run_kwargs[task_input.name]
                )
                continue
        return task.run(Session(shared_ctx=shared_ctx, root_group=self))

    def _show_task_info(self, task: AnyTask):
        description = task.description
        inputs = task.inputs
        if description != task.name and description != "":
            print(stylize_section_header("DESCRIPTION"))
            print(description)
            print()
        if len(inputs) > 0:
            print(stylize_section_header("INPUTS"))
            max_input_name_length = max(len(task_input.name) for task_input in inputs)
            for task_input in inputs:
                task_input_name = task_input.name.ljust(max_input_name_length + 1)
                print(f"  --{task_input_name}: {task_input.description}")
            print()

    def _show_group_info(self, group: AnyGroup):
        if group.banner != "":
            print(group.banner)
            print()
        if group.description != group.name and group.description != "":
            print(stylize_section_header("DESCRIPTION"))
            print(group.description)
            print()
        subgroups = get_non_empty_subgroups(group)
        if len(subgroups) > 0:
            print(stylize_section_header("GROUPS"))
            max_subgroup_alias_length = max(len(s) for s in subgroups)
            for alias, subgroup in subgroups.items():
                alias = alias.ljust(max_subgroup_alias_length + 1)
                print(f"  {alias}: {subgroup.description}")
            print()
        subtasks = get_subtasks(group)
        if len(subtasks) > 0:
            print(stylize_section_header("TASKS"))
            max_subtask_alias_length = max(len(s) for s in subtasks)
            for alias, subtask in subtasks.items():
                alias = alias.ljust(max_subtask_alias_length + 1)
                print(f"  {alias}: {subtask.description}")
            print()

    def _extract_kwargs_from_args(
        self, args: list[str]
    ) -> tuple[dict[str, Any], list[str]]:
        residual_args = []  # To store positional arguments
        kwargs = {}  # To store options as a dictionary
        i = 0
        while i < len(args):
            arg = args[i]
            if arg.startswith("--"):
                # Handle key-value pairs like --keyword=value
                if "=" in arg:
                    key, value = arg[2:].split("=", 1)
                    kwargs[key] = value
                else:
                    # Handle flags like --this followed by a value or set to True
                    key = arg[2:]
                    # Check if the next item is a value or another flag
                    if i + 1 < len(args) and not args[i + 1].startswith("-"):
                        kwargs[key] = args[i + 1]
                        i += 1  # Skip the next argument as it's a value
                    else:
                        kwargs[key] = True
            elif arg.startswith("-"):
                # Handle short flags like -t or -n
                key = arg[1:]
                kwargs[key] = True
            else:
                # Anything else is considered a positional argument
                residual_args.append(arg)
            i += 1
        return kwargs, residual_args


cli = Cli()


@make_task(name="version", description="🌟 Get current version", retries=0, group=cli)
def get_version(_: AnyContext):
    return CFG.VERSION


server_group = cli.add_group(
    Group(name="server", description="🌐 Server related command")
)


@make_task(
    name="start-server",
    description="🚀 Start Zrb Web Server",
    cli_only=True,
    retries=0,
    group=server_group,
    alias="start",
)
async def start_server(_: AnyContext):
    from uvicorn import Config, Server

    app = create_web_app(cli, web_config, session_state_logger)
    server = Server(
        Config(
            app=app,
            host="0.0.0.0",
            port=CFG.WEB_HTTP_PORT,
            loop="asyncio",
        )
    )
    await server.serve()



from zrb.builtin.base64 import decode_base64, encode_base64, validate_base64
from zrb.builtin.git import (
    get_git_diff,
    git_commit,
    git_pull,
    git_push,
    prune_local_branches,
)
from zrb.builtin.git_subtree import git_add_subtree, git_pull_subtree, git_push_subtree
from zrb.builtin.http import generate_curl, http_request
from zrb.builtin.jwt import decode_jwt, encode_jwt, validate_jwt
from zrb.builtin.llm.llm_ask import llm_ask
from zrb.builtin.md5 import hash_md5, sum_md5, validate_md5
from zrb.builtin.project.add.fastapp.fastapp_task import add_fastapp_to_project
from zrb.builtin.project.create.project_task import create_project
from zrb.builtin.python import format_python_code
from zrb.builtin.random import shuffle_values, throw_dice
from zrb.builtin.setup.asdf.asdf import setup_asdf
from zrb.builtin.setup.latex.ubuntu import setup_latex_on_ubuntu
from zrb.builtin.setup.tmux.tmux import setup_tmux
from zrb.builtin.setup.ubuntu import setup_ubuntu
from zrb.builtin.setup.zsh.zsh import setup_zsh
from zrb.builtin.shell.autocomplete.bash import make_bash_autocomplete
from zrb.builtin.shell.autocomplete.subcmd import get_shell_subcommands
from zrb.builtin.shell.autocomplete.zsh import make_zsh_autocomplete
from zrb.builtin.todo import (
    add_todo,
    archive_todo,
    complete_todo,
    edit_todo,
    list_todo,
    log_todo,
    show_todo,
)
from zrb.builtin.uuid import (
    generate_uuid_v1,
    generate_uuid_v3,
    generate_uuid_v4,
    generate_uuid_v5,
    validate_uuid,
    validate_uuid_v1,
    validate_uuid_v3,
    validate_uuid_v4,
    validate_uuid_v5,
)

assert create_project
assert add_fastapp_to_project
assert get_shell_subcommands
assert make_bash_autocomplete
assert make_zsh_autocomplete
assert encode_base64
assert decode_base64
assert validate_base64
assert encode_jwt
assert decode_jwt
assert validate_jwt
assert llm_ask
assert hash_md5
assert sum_md5
assert validate_md5
assert get_git_diff
assert prune_local_branches
assert format_python_code
assert git_commit
assert git_pull
assert git_push
assert git_add_subtree
assert git_pull_subtree
assert git_push_subtree
assert list_todo
assert add_todo
assert archive_todo
assert edit_todo
assert complete_todo
assert log_todo
assert show_todo
assert throw_dice
assert shuffle_values
assert setup_ubuntu
assert setup_latex_on_ubuntu
assert setup_asdf
assert setup_tmux
assert setup_zsh
assert validate_uuid
assert validate_uuid_v1
assert validate_uuid_v3
assert validate_uuid_v4
assert validate_uuid_v5
assert generate_uuid_v1
assert generate_uuid_v3
assert generate_uuid_v4
assert generate_uuid_v5
assert http_request
assert generate_curl



import json
from typing import TYPE_CHECKING, Any

from pydantic import BaseModel

from zrb.attr.type import BoolAttr, IntAttr
from zrb.context.any_context import AnyContext
from zrb.llm_config import llm_config
from zrb.task.llm.agent import run_agent_iteration
from zrb.task.llm.history import (
    count_part_in_history_list,
    replace_system_prompt_in_history_list,
)
from zrb.task.llm.typing import ListOfDict
from zrb.util.attr import get_bool_attr, get_int_attr
from zrb.util.cli.style import stylize_faint

if TYPE_CHECKING:
    from pydantic_ai.models import Model
    from pydantic_ai.settings import ModelSettings
else:
    Model = Any
    ModelSettings = Any


def get_history_summarization_threshold(
    ctx: AnyContext,
    history_summarization_threshold_attr: IntAttr | None,
    render_history_summarization_threshold: bool,
) -> int:
    """Gets the history summarization threshold, handling defaults and errors."""
    try:
        return get_int_attr(
            ctx,
            history_summarization_threshold_attr,
            # Use llm_config default if attribute is None
            llm_config.default_history_summarization_threshold,
            auto_render=render_history_summarization_threshold,
        )
    except ValueError as e:
        ctx.log_warning(
            f"Could not convert history_summarization_threshold to int: {e}. "
            "Defaulting to -1 (no threshold)."
        )
        return -1


def should_summarize_history(
    ctx: AnyContext,
    history_list: ListOfDict,
    should_summarize_history_attr: BoolAttr | None,  # Allow None
    render_summarize_history: bool,
    history_summarization_threshold_attr: IntAttr | None,  # Allow None
    render_history_summarization_threshold: bool,
) -> bool:
    """Determines if history summarization should occur based on length and config."""
    history_part_count = count_part_in_history_list(history_list)
    if history_part_count == 0:
        return False
    summarization_threshold = get_history_summarization_threshold(
        ctx,
        history_summarization_threshold_attr,
        render_history_summarization_threshold,
    )
    if summarization_threshold == -1 or summarization_threshold > history_part_count:
        return False
    return get_bool_attr(
        ctx,
        should_summarize_history_attr,
        # Use llm_config default if attribute is None
        llm_config.default_summarize_history,
        auto_render=render_summarize_history,
    )


class SummarizationConfig(BaseModel):
    model_config = {"arbitrary_types_allowed": True}
    model: Model | str | None = None
    settings: ModelSettings | None = None
    prompt: str
    retries: int = 1


async def summarize_history(
    ctx: AnyContext,
    config: SummarizationConfig,
    conversation_context: dict[str, Any],
    history_list: ListOfDict,
) -> dict[str, Any]:
    """Runs an LLM call to summarize history and update the context."""
    from pydantic_ai import Agent

    ctx.log_info("Attempting to summarize conversation history...")

    summarization_agent = Agent(
        model=config.model,
        system_prompt=config.prompt,
        tools=[],  # No tools needed for summarization
        mcp_servers=[],
        model_settings=config.settings,
        retries=config.retries,
    )

    # Prepare context and history for summarization prompt
    try:
        context_json = json.dumps(conversation_context)
        history_to_summarize_json = json.dumps(history_list)
        summarization_user_prompt = "\n".join(
            [
                f"Current Context: {context_json}",
                f"Conversation History to Summarize: {history_to_summarize_json}",
            ]
        )
    except Exception as e:
        ctx.log_warning(f"Error formatting context/history for summarization: {e}")
        return conversation_context  # Return original context if formatting fails

    try:
        ctx.print(stylize_faint("[Summarization Triggered]"), plain=True)
        summary_run = await run_agent_iteration(
            ctx=ctx,
            agent=summarization_agent,
            user_prompt=summarization_user_prompt,
            history_list=[],  # Summarization agent doesn't need prior history
        )
        if summary_run and summary_run.result.output:
            summary_text = str(summary_run.result.output)
            usage = summary_run.result.usage()
            ctx.print(stylize_faint(f"[Token Usage] {usage}"), plain=True)
            # Update context with the new summary
            conversation_context["history_summary"] = summary_text
            ctx.log_info("History summarized and added/updated in context.")
            ctx.log_info(f"Conversation summary: {summary_text}")
        else:
            ctx.log_warning("History summarization failed or returned no data.")
    except Exception as e:
        ctx.log_warning(f"Error during history summarization: {e}")

    return conversation_context


async def maybe_summarize_history(
    ctx: AnyContext,
    history_list: ListOfDict,
    conversation_context: dict[str, Any],
    should_summarize_history_attr: BoolAttr | None,  # Allow None
    render_summarize_history: bool,
    history_summarization_threshold_attr: IntAttr | None,  # Allow None
    render_history_summarization_threshold: bool,
    model: str | Model | None,
    model_settings: ModelSettings | None,
    summarization_prompt: str,
) -> tuple[ListOfDict, dict[str, Any]]:
    """Summarizes history and updates context if enabled and threshold met."""
    shorten_history_list = replace_system_prompt_in_history_list(history_list)
    if should_summarize_history(
        ctx,
        shorten_history_list,
        should_summarize_history_attr,
        render_summarize_history,
        history_summarization_threshold_attr,
        render_history_summarization_threshold,
    ):
        # Use summarize_history defined above
        updated_context = await summarize_history(
            ctx=ctx,
            config=SummarizationConfig(
                model=model,
                settings=model_settings,
                prompt=summarization_prompt,
            ),
            conversation_context=conversation_context,
            history_list=shorten_history_list,  # Pass the full list for context
        )
        # Truncate the history list after summarization
        return [], updated_context
    return history_list, conversation_context



import os

from my_app_name._zrb.column.add_column_task import add_my_app_name_column
from my_app_name._zrb.config import (
    ACTIVATE_VENV_SCRIPT,
    APP_DIR,
    MONOLITH_ENV_VARS,
    TEST_ENV_VARS,
)
from my_app_name._zrb.entity.add_entity_task import add_my_app_name_entity
from my_app_name._zrb.format_task import format_my_app_name_code
from my_app_name._zrb.group import (
    app_create_migration_group,
    app_group,
    app_migrate_group,
    app_run_group,
)
from my_app_name._zrb.input import run_env_input
from my_app_name._zrb.module.add_module_task import add_my_app_name_module
from my_app_name._zrb.task_util import (
    create_migration,
    migrate_module,
    run_microservice,
    run_my_app_name,
)
from my_app_name._zrb.venv_task import prepare_venv

from zrb import AnyContext, CmdPath, CmdTask, EnvFile, EnvMap, Task, make_task

assert add_my_app_name_entity
assert add_my_app_name_module
assert add_my_app_name_column
assert format_my_app_name_code

# 🧪 Test ======================================================================


@make_task(
    name="prepare-my-app-name-test",
)
def prepare_test(ctx: AnyContext):
    db_test_path = os.path.join(APP_DIR, "test.db")
    if os.path.exists(db_test_path):
        ctx.print(f"Removing test db: {db_test_path}")
        os.remove(db_test_path)


migrate_test = Task(name="migrate-test")

test_app = app_group.add_task(
    CmdTask(
        name="test-my-app-name",
        description="🧪 Test My App Name",
        env=EnvMap(vars=TEST_ENV_VARS),
        cwd=APP_DIR,
        cmd=CmdPath(os.path.join(APP_DIR, "test.sh")),
        retries=0,
    ),
    alias="test",
)
prepare_venv >> prepare_test >> migrate_test >> test_app


# 🚀 Run/Migrate All ===========================================================

run_all = app_run_group.add_task(
    Task(
        name="run-my-app-name",
        description="🟢 Run My App Name as monolith and microservices",
    ),
    alias="all",
)

migrate_all = app_migrate_group.add_task(
    Task(
        name="migrate-my-app-name",
        description="📦 Run My App Name DB migration for monolith and microservices",
    ),
    alias="all",
)

create_all_migration = app_create_migration_group.add_task(
    Task(
        name="create-my-app-name-migration",
        description="📦 Create My App Name DB migration",
    ),
    alias="all",
)

# 🗿 Run/Migrate Monolith =====================================================

run_monolith = app_run_group.add_task(
    CmdTask(
        name="run-monolith-my-app-name",
        description="🗿 Run My App Name as a monolith",
        input=run_env_input,
        env=[
            EnvFile(path=os.path.join(APP_DIR, "template.env")),
            EnvMap(vars=MONOLITH_ENV_VARS),
        ],
        cwd=APP_DIR,
        cmd=[
            ACTIVATE_VENV_SCRIPT,
            run_my_app_name,
        ],
        render_cmd=False,
        retries=2,
    ),
    alias="monolith",
)
prepare_venv >> run_monolith >> run_all

migrate_monolith = app_migrate_group.add_task(
    Task(
        name="migrate-monolith-my-app-name",
        description="🗿 Run My App Name DB migration for monolith",
    ),
    alias="monolith",
)
migrate_monolith >> migrate_all

# 🌐 Run/Migrate Microsevices ==================================================

run_microservices = app_run_group.add_task(
    Task(
        name="run-microservices-my-app-name",
        description="🌐 Run My App Name as microservices",
    ),
    alias="microservices",
)
run_microservices >> run_all

migrate_microservices = app_migrate_group.add_task(
    Task(
        name="migrate-microservices-my-app-name",
        description="🌐 Run My App Name DB migration for microservices",
    ),
    alias="microservices",
)
migrate_microservices >> migrate_all

# 📡 Run/Migrate Gateway =======================================================

run_gateway = app_run_group.add_task(
    run_microservice("gateway", 3001), alias="svc-gateway"
)
prepare_venv >> run_gateway >> run_microservices

create_gateway_migration = app_create_migration_group.add_task(
    create_migration("gateway"), alias="gateway"
)
prepare_venv >> create_gateway_migration >> create_all_migration

migrate_monolith_gateway = migrate_module("gateway", as_microservices=False)
prepare_venv >> migrate_monolith_gateway >> [migrate_monolith, run_monolith]

migrate_microservices_gateway = app_migrate_group.add_task(
    migrate_module("gateway", as_microservices=True),
    alias="svc-gateway",
)
prepare_venv >> migrate_microservices_gateway >> [migrate_microservices, run_gateway]

migrate_test_gateway = migrate_module(
    "gateway", as_microservices=False, additional_env_vars=TEST_ENV_VARS
)
prepare_venv >> migrate_test_gateway >> migrate_test


# 🔐 Run/Migrate Auth ==========================================================

run_auth = app_run_group.add_task(run_microservice("auth", 3002), alias="svc-auth")
prepare_venv >> run_auth >> run_microservices

create_auth_migration = app_create_migration_group.add_task(
    create_migration("auth"), alias="auth"
)
prepare_venv >> create_auth_migration >> create_all_migration

migrate_monolith_auth = migrate_module("auth", as_microservices=False)
prepare_venv >> migrate_monolith_auth >> [migrate_monolith, run_monolith]

migrate_microservices_auth = app_migrate_group.add_task(
    migrate_module("auth", as_microservices=True), alias="svc-auth"
)
prepare_venv >> migrate_microservices_auth >> [migrate_microservices, run_auth]

migrate_test_auth = migrate_module(
    "auth", as_microservices=False, additional_env_vars=TEST_ENV_VARS
)
prepare_venv >> migrate_test_auth >> migrate_test
